/* SimpleCapi.cpp

implementation of the SimpleCapi class. (See SimpleCapi.h)

Part of the caiviar package. 

Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#ifdef LINUX
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "CapiHandler2.h"
#include "Log.h"
#include "SimpleCapi.h"
#include "wav2law.h"
#define EnterCriticalSection pthread_mutex_lock
#define LeaveCriticalSection pthread_mutex_unlock
#define stricmp strcasecmp
#else
#include "CapiHandler2.h"
#include "log.h"
#include "SimpleCapi.h"
#include <conio.h>
#include "wav2law.h"
#endif

#define CAPI ((CCapiHandler2*)(this->internal))
#define LOG ((((CCapiHandler2*)(this->internal))->log))

#ifdef LINUX
pthread_mutex_t line_mutex[64];
#else
CRITICAL_SECTION line_mutex[64];
#endif

int instancecount = 0;

static void readline(FILE*fi, char*line, int maxlen) 
{
    int pos = 0;
    while(!feof(fi)) {
        if(!fread(&line[pos],1,1,fi))
            break;
        if(line[pos] == 13 || line[pos]==10)
            break;
	if(pos<maxlen-1) 
	    pos++;
    }
    line[pos]=0;

    /* cut off comments */
    char*x=strchr(line,'#');
    if(x) {
        *x=0;
	pos = x-line;
    }

    /* cut off trailing whitespace */
    while(pos>=1 && (line[pos-1]==32 || line[pos-1]==9)) {
	pos--;
	line[pos]=0;
    }
}

static char* properties_msisdn = 0;
static char* properties_controller = 0; // all controllers

SimpleCapi::SimpleCapi(char*propertyfile, Logger*log) 
{
    char line[256]; // notice that line doesn't have anything to do with isdn, here.;)
    FILE*fi=0;
   
    if(propertyfile)
	fi = fopen(propertyfile,"rb");

    if(!log) {
	log = new Logger();
	log->initLog (0,-1,0,0,-1,2); //-1)?
    }

    int properties_allcalls = 0;
    if(!fi && propertyfile) {
        log->logf("<warning> Couldn't open property file \"%s\".\n", propertyfile);
    }
    while(fi && !feof(fi)) {
        readline(fi,line, 256);
        char*a=strchr(line, '=');
        if(a) {
            *a=0;
            if (!stricmp(line, "local.msisdn"))
                properties_msisdn=strdup(a+1);
            else if (!stricmp(line, "local.allcalls"))
                properties_allcalls=atoi(a+1);
            else if (!stricmp(line, "local.controller"))
                properties_controller=strdup(a+1);
        }
    }
    if(fi)
	fclose(fi);

    this->internal = (void*)(new CCapiHandler2(log));
    ((CCapiHandler2*)(this->internal))->allcalls_hack = properties_allcalls;
    instancecount++;
    if(instancecount>=2) {
        LOG->logf("<error> Only one instance of SimpleCapi should be created.");
    }
}

SimpleCapi::~SimpleCapi() 
{
    LOG->log("<verbose> SimpleCapi::~SimpleCapi()");
    Logger*log = LOG;
    delete CAPI;
    delete log;
    instancecount--;
}

int SimpleCapi::open (char*msisdn) 
{
    LOG->logf("<verbose> SimpleCapi::open(%s)",msisdn);
    if(!msisdn || !*msisdn) {
        msisdn = properties_msisdn;
	LOG->logf("<verbose> SimpleCapi::open(): using msisdn \"%s\" and controllermask [%s] from properties", msisdn,
		properties_controller?properties_controller:"all");
    } else {
	msisdn = strdup(msisdn);
    }
    EnterCriticalSection(&CAPI->capi_mutex); //probably not needed.
    int retval = CAPI->open(msisdn, properties_controller);
    LeaveCriticalSection(&CAPI->capi_mutex);
#ifndef LINUX
    if(retval>=0)
        InitializeCriticalSection(&line_mutex[retval]);
#endif
    LOG->logf("<verbose> SimpleCapi::open() returns line %d", retval);
    return retval;
}

void SimpleCapi::close (int line) 
{
    LOG->logf("<verbose> SimpleCapi::close(%d)",line);
    if(!CAPI->checkLine(line))
        return;
    EnterCriticalSection(&line_mutex[line]);
    CAPI->close(line);
    LeaveCriticalSection(&line_mutex[line]);
#ifndef LINUX
    DeleteCriticalSection(&line_mutex[line]);
#endif
}

void SimpleCapi::acceptCalls (int line) 
{
    LOG->logf("<verbose> SimpleCapi::acceptCalls(%d)",line);
    if(!CAPI->checkLine(line))
        return;
    EnterCriticalSection(&line_mutex[line]);
    CAPI->acceptCalls(line);
    LeaveCriticalSection(&line_mutex[line]);
}

int SimpleCapi::call (int line, char* msisdn, int timeout) 
{
    LOG->logf("<verbose> SimpleCapi::call(%d, %s, %d)",line,msisdn,timeout);
    if(!CAPI->checkLine(line))
        return 0;
    EnterCriticalSection(&line_mutex[line]);
    CAPI->call (line, msisdn, -1);
    LeaveCriticalSection(&line_mutex[line]);

    float timeout2=(float)timeout;
    while(1) {
        EnterCriticalSection(&line_mutex[line]);
        ISDNEvent event = CAPI->listen(line, timeout2);
        LeaveCriticalSection(&line_mutex[line]);
        if(event.type == CAPI_EVENT_CONNECTED) {
	    break;
        }
        /* CapiHandler2 will return an HANGUP event, even
         * if there hasn't been a connection yet. this means
         * for some reason we didn't manage to call the other
         * party
         */
        if(event.type == CAPI_EVENT_HANGUP) {
            if(event.reason)
                return event.reason;
            else
                return CAPI_REASON_TECHNICAL;
        }
        if(event.type == TIMEOUT) {
            return CAPI_REASON_NOANSWER;
        }
    }
    return 0;
}

int SimpleCapi::call_nonblock (int line, char* msisdn) 
{
    LOG->logf("<verbose> SimpleCapi::call_nonblock(%d, %s)",line, msisdn);
    if(!CAPI->checkLine(line))
        return -1;
    EnterCriticalSection(&line_mutex[line]);
    int retval  = CAPI->call (line, msisdn, -1);
    LeaveCriticalSection(&line_mutex[line]);

    return retval;
}

int SimpleCapi::play (int line, const char* filename) 
{
    FILE*fi = fopen (filename,"rb");
    if(!fi) {
        LOG->logf("<error> Could not find file\"%s\"!", filename);
        return -1;
    }
    fseek (fi, 0, SEEK_END);
    long int size = ftell(fi);
    fseek (fi, 0, SEEK_SET);

    LOG->logf ("<verbose> SimpleCapi::play(%d, %s) (%d bytes)", line, filename, size);

    char*data = (char*)malloc (size);
    fread (data, 1, size, fi);
    fclose (fi);
    int playjob = play ( line, data, size);
    free (data);
    return playjob;
}

int SimpleCapi::play (int line, void*mem, int len) 
{
    LOG->logf ("<verbose> SimpleCapi::play(%d, ptr, %d)", line,len);
    int riff = 0;
    unsigned char*mem2 = (unsigned char*)mem;
    if(mem2[0] == 0x52 && mem2[1] == 0x49 && mem2[2] == 0x46 && mem2[3] == 0x46) //RIFF
    {
        LOG->log ("<notice> Performing wav to law conversion");
        mem = wav2law(mem2, len, &len);
        riff = 1;
    }
    if(!CAPI->checkLine(line))
        return -1;
    EnterCriticalSection(&line_mutex[line]);
    int ret=CAPI->play (line, mem, len);
    LeaveCriticalSection(&line_mutex[line]);
    if(riff)
        free(mem);
    return ret;
}

int SimpleCapi::record (int line, void*buf, int bufsize) 
{
    LOG->logf ("<verbose> SimpleCapi::record(%d, ptr, %d)",line,bufsize);
    if(!CAPI->checkLine(line))
        return -1;
    EnterCriticalSection(&line_mutex[line]);
    int ret = CAPI->record(line,buf,bufsize);
    LeaveCriticalSection(&line_mutex[line]);

    return ret;
}

int SimpleCapi::getRecordSize (int line) 
{
    EnterCriticalSection(&line_mutex[line]);
    int ret = CAPI->getRecordSize(line);
    LeaveCriticalSection(&line_mutex[line]);
    return ret;
}

int SimpleCapi::fax (int line, char*msisdn, char*filename) 
{
    int faxformat = 5; //ascii
    {
        FILE*fi= fopen(filename, "rb");
        char a[4];
        if(fi && fread(a,1,4,fi)==4) {
            if (!strncmp(a,"Sfff",4))
                faxformat = 0;
	    fclose(fi);
        }
    }

    LOG->logf ("<verbose> SimpleCapi::fax(%d, %s, %s) (faxformat = %d)", line,msisdn,filename, faxformat);
    EnterCriticalSection(&line_mutex[line]);
    if(!CAPI->checkLine(line))
        return 0;
    int ret =  CAPI->call (line, msisdn, faxformat);
    LeaveCriticalSection(&line_mutex[line]);
    if (!ret) {
        printf("no answer\n");
        return 0;
    }
    int done=0;
    while (!done) {
        ISDNEvent event = listen(line, 0);
        switch (event.type) {
        case CAPI_EVENT_CONNECTED:
            done = 1;
            break;
        case CAPI_EVENT_HANGUP:
	    return 0;
        default:
            printf ("Unexpected: %s\n", eventToString (&event));
            break;
        }
    }

    /* to "play" means to transfer data, which is exactly what we want to
     * do here */
    int playjob = play (line, filename);
    done=0;
    while (!done) {
        ISDNEvent event = listen(line, 0);
        switch (event.type) {
        case CAPI_EVENT_HANGUP:
            printf ("hung up\n");
            return 0;

        case CAPI_EVENT_DONE:
            if (event.jobid == playjob) {
                done = 1;
            }
            break;

        default:
            printf ("Unexpected: %s\n", eventToString (&event));
            break;
        }
    }

    done = 0;
    while(!done) {
        ISDNEvent event = listen(line, 1);
        if(event.type == CAPI_EVENT_HANGUP) {
            done=1;
        }
    }

    return 1;
}

int SimpleCapi::callX31 (int line, char* msn, char* datex) 
{
    LOG->logf("<verbose> SimpleCapi::sendX31data(%d, %s)", line, msn);

    EnterCriticalSection(&line_mutex[line]);
    if (!CAPI->checkLine(line)) {
        return -1;
    }
    
    CAPI->setDatexAddress(line, datex);
    int ret = CAPI->call(line, msn, 100); // start X25 DTE-DTE call
    LeaveCriticalSection(&line_mutex[line]);
    if (!ret) {
        printf("no answer");
        return 0;
    }

    int done = 0;
    while (!done) {
        ISDNEvent event = listen(line, 0);
        switch (event.type) {
        case CAPI_EVENT_CONNECTED:
            done = 1;
            break;
        default:
            printf("Unexpected: %s\n", eventToString(&event));
	    return -2;
            break;
        }
    }
    return 1;
}

void SimpleCapi::stop (int jobid) 
{
    LOG->logf("<verbose> SimpleCapi::stop(%d)", jobid);
    EnterCriticalSection(&CAPI->capi_mutex); //FIXME
    CAPI->stop(jobid);
    LeaveCriticalSection(&CAPI->capi_mutex);
}

ISDNEvent SimpleCapi::listen (int line, int timeout) 
{
    float timeout2 = (float)timeout;
    LOG->logf("<verbose> SimpleCapi::listen(%d, %f)", line, timeout2);
    ISDNEvent ret;
    ret.type = -1;
    if(!CAPI->checkLine(line))
        return ret;
    EnterCriticalSection(&line_mutex[line]);
    ret = CAPI->listen(line, timeout2);
    LeaveCriticalSection(&line_mutex[line]);
    return ret;
}

int SimpleCapi::answer (ISDNEvent*event) 
{
    int line = event->line;
    LOG->logf("<verbose> SimpleCapi::answer(%s) (line %d)", eventToString(event), line);
    if(!CAPI->checkLine(line))
        return 0;
    EnterCriticalSection(&line_mutex[line]);
    int ret = CAPI->answer(event);
    LeaveCriticalSection(&line_mutex[line]);
    return ret;
}

/* perform several supplemental services. 
   hold and retrieve should now be used through
   SimpleCapi::hold and SimpleCapi::retrieve.
   ect (explicit call transfer) and forwarding
   are untested and probably not working */
int SimpleCapi::supplemental (int line, char* command) 
{
    LOG->logf("<verbose> SimpleCapi::supplemental(%d, %s)", line, command);
    if(!CAPI->checkLine(line))
        return 0;
    EnterCriticalSection(&line_mutex[line]);
    int ret;

    if(!strcmp(command, "hold"))
        ret = CAPI->hold(line);
    else if(!strcmp(command, "retrieve"))
        ret = CAPI->retrieve(line);
    else if(!strcmp(command, "3pty"))
        ret = CAPI->conference(line);
    else if(!strcmp(command, "ect"))
        ret = CAPI->ect(line);
    else
        ret = CAPI->forward(line, command);

    LeaveCriticalSection(&line_mutex[line]);
    return ret;
}

int SimpleCapi::hold (int line)
{
    LOG->logf("<verbose> SimpleCapi::hold(%d)", line);
    int ret, jobid;
    jobid = this->supplemental(line, "hold");
    LOG->logf("<verbose> SimpleCapi::hold jobid:%d", jobid);

    if(!jobid)
	return 0;
    ISDNEvent event;
    while(1)
    {
	event = this->listen(line, 10);
	switch(event.type)
	{
	    case TIMEOUT:
		return 0;
	    case CAPI_EVENT_HANGUP:
	    /* (!) we return 1 here, even though the hold didn't succeed. 
		   however the line is free and it's upon the subsequent 
		   retrieve() to return an error. */
		return 1;
	    case CAPI_EVENT_DONE:
		if(jobid == event.jobid) {
		    /* big hack - as CapiHandler needs some extra time to
		       complete the hold() procedure, we call listen again
		       and hope the event is a TIMEOUT */
		    this->listen(line, 1);
		    return 1;
		}
		LOG->logf("<verbose> Warning: job %d done while waiting for hold().", event.jobid);
	    break;
	}
    }
}

int SimpleCapi::retrieve (int line)
{
    LOG->logf("<verbose> SimpleCapi::hold(%d)", line);
    int ret, jobid;
    jobid = this->supplemental(line, "retrieve");
    if(!jobid)
	return 0;
    ISDNEvent event;
    while(1)
    {
	event = this->listen(line, 10);
	switch(event.type)
	{
	    case TIMEOUT:
		return 0;
	    case CAPI_EVENT_HANGUP:
		return 0;
	    case CAPI_EVENT_DONE:
		if(jobid == event.jobid) {
		    /* big hack - as CapiHandler needs some extra time to
		       complete the retrieve() procedure, we call listen again
		       and hope the event is a TIMEOUT */
		    this->listen(line, 1);
		    return 1;
		}
		LOG->logf("<verbose> Warning: job %d done while waiting for hold().", jobid);
	    break;
	}
    }
}

int SimpleCapi::forward (int line, char*msisdn) 
{
    LOG->logf("<verbose> SimpleCapi::forward(%d, %s)", line, msisdn);
    LOG->logf("<warning> forward() is deprecated. Please use hold(), retrieve() and supplemental().");
    return this->supplemental (line, msisdn);
}

void SimpleCapi::hangup (int line) 
{
    LOG->logf ("<verbose> SimpleCapi::hangup(%d)", line);
    if(!CAPI->checkLine(line))
        return ;
    EnterCriticalSection(&line_mutex[line]);
    CAPI->hangup(line);
    LeaveCriticalSection(&line_mutex[line]);
}

char* SimpleCapi::eventToString (ISDNEvent*event) 
{
    LOG->logf("<verbose> SimpleCapi::eventToString(line:%d)", event->line);
    char* ret = CAPI->eventToString(event);
    return ret;
}

